<?php

defined('BASEPATH') OR exit('No direct script access allowed');
/**
 * Pancake
 *
 * A simple, fast, self-hosted invoicing application
 *
 * @package		Pancake
 * @author		Pancake Dev Team
 * @copyright           Copyright (c) 2014, Pancake Payments
 * @license		http://pancakeapp.com/license
 * @link		http://pancakeapp.com
 * @since		Version 1.0
 */
// ------------------------------------------------------------------------

/**
 * The install wizard controller
 *
 * @subpackage	Controllers
 * @category	Wizard
 */
class Wizard extends CI_Controller {

    /**
     * @var	string	The path to the config folder
     */
    private $config_path;

    /**
     * @var	string	The path to the config.php file
     */
    private $config_file;

    /**
     * @var	string	The path to the uploads folder
     */
    private $upload_path;

    // ------------------------------------------------------------------------

    public function __construct() {
        parent::__construct();
        $this->config_path = FCPATH . 'system/pancake/config';
        $this->config_file = $this->config_path . "/config.php";
        $this->upload_path = FCPATH . 'uploads/';

        include_once FCPATH . 'system/pancake/libraries/Asset.php';
        Asset::add_path(APPPATH . 'assets/');
        Asset::set_asset_url(BASE_URL);
    }

    public function _detect_url_rewriting() {
        include_once APPPATH . 'libraries/HTTP_Request.php';
        $http = new HTTP_Request();
        try {
            $url_without_index_php = BASE_URL . "admin";
            $contents = $http->request($url_without_index_php);
            if (stristr($contents, "run the installer") === false) {
                $contents = $http->request(BASE_URL . "index.php/admin");
                if (stristr($contents, "run the installer") === false) {
                    $contents = $http->request(BASE_URL . "index.php?/admin");
                    if (stristr($contents, "run the installer") === false) {
                        # Could not detect any valid way to access Pancake, so assume it works if there's a .htaccess and it's Apache, otherwise default to "index.php".

                        $htaccess_servers = array(
                            "Apache",
                            "LiteSpeed",
                            "WebServerX",
                            "1984"
                        );

                        $use_index_php = "index.php";
                        foreach ($htaccess_servers as $server) {
                            if (strpos($_SERVER["SERVER_SOFTWARE"], $server) !== false) {
                                $use_index_php = is_file(FCPATH . '.htaccess') ? '' : 'index.php';
                            }
                        }
                    } else {
                        # Found "run the installer", works with index.php?.
                        $use_index_php = "index.php?";
                    }

                } else {
                    # Found "run the installer", works with index.php.
                    $use_index_php = "index.php";
                }
            } else {
                # Found "run the installer", works without index.php.
                $use_index_php = "";
            }
        } catch (Exception $e) {
            # Could not detect URL Rewriting correctly; defaulting to "index.php?".
            return 'index.php?';
        }

        if ($use_index_php !== $this->config->item("index_page")) {
            if (is_really_writable($this->config_file)) {
                file_put_contents($this->config_file, str_replace("\$config['index_page'] = \$config['index_page'];", "\$config['index_page'] = '$use_index_php'; #CORRECTED", file_get_contents($this->config_file)));
                if (strstr(file_get_contents($this->config_file), "#CORRECTED") !== false) {
                    # Everything is OK; the file was corrected.
                    return true;
                } else {
                    # Something went wrong while updating the file.
                    return false;
                }
            } else {
                # The file can't be corrected.
                return false;
            }
        } else {
            # Everything is OK; detection was correct.
            return true;
        }
    }

    public function index() {
        $this->_detect_url_rewriting();

        $this->session->set_userdata('last_step', '0');
        $this->_output_step('steps/welcome');
    }

    public function step1() {
        $data = array();
        $data['is_url_rewriting_working'] = $this->_detect_url_rewriting();
        $data['config_writable'] = is_really_writable($this->config_path) ? TRUE : FALSE;
        $data['upload_writable'] = is_really_writable($this->upload_path) ? TRUE : FALSE;
        $data['curl_installed'] = function_exists('curl_init');
        $data['php_version'] = (version_compare(PHP_VERSION, "5.3.0", "gt"));

        $data['installed'] = array(
            'gd' => (extension_loaded('gd') && function_exists('gd_info')),
            'json' => (extension_loaded('json') && function_exists('json_encode')),
            'dom' => (extension_loaded('dom')),
            'xml' => (extension_loaded('xml')),
            'mysql' => (extension_loaded('mysqli') || extension_loaded('mysql'))
        );

        include_once APPPATH . 'libraries/HTTP_Request.php';
        $http = new HTTP_Request();
        try {
            $http->request(MANAGE_PANCAKE_BASE_URL);
            $data['manage_pancakeapp'] = true;
        } catch (Exception $e) {
            $data['manage_pancakeapp'] = false;
        }

        $data['can_continue'] = true;
        foreach ($data['installed'] as $installed) {
            if (!$installed) {
                $data['can_continue'] = false;
            }
        }

        $data['can_continue'] = ($data['can_continue'] && $data['is_url_rewriting_working'] && $data['manage_pancakeapp'] && $data['config_writable'] AND $data['upload_writable'] AND $data['php_version'] AND $data['curl_installed']);
        if ($data['can_continue']) {
            $this->session->set_userdata('last_step', '1');
        }
        $this->_output_step('steps/step1', $data);
    }

    public function step2() {
        if ($this->session->userdata('last_step') != '1') {
            redirect('wizard/step1');
        }
        $data['hostname'] = $this->input->post('hostname');
        $data['database'] = $this->input->post('database');
        $data['dbprefix'] = $this->input->post('dbprefix') ? $this->input->post('dbprefix') : 'pancake_';
        $data['username'] = $this->input->post('username');
        $data['password'] = $this->input->post('password');
        $data['port'] = $this->input->post('port');
        if ($_POST) {
            $data['port'] = (!$data['port']) ? '3306' : $data['port'];

            $link = @mysql_connect($data['hostname'] . ':' . $data['port'], $data['username'], $data['password'], TRUE);

            if ($link) {
                // If the database is not there create it
                mysql_query('CREATE DATABASE IF NOT EXISTS ' . $data['database'], $link);
            }

            if ($link AND @ mysql_select_db($data['database'], $link)) {
                $this->session->set_userdata(array(
                    'hostname' => $data['hostname'],
                    'database' => $data['database'],
                    'dbprefix' => $data['dbprefix'],
                    'mysql_username' => $data['username'],
                    'password' => $data['password'],
                    'port' => $data['port'],
                    'last_step' => 2,
                ));
                redirect('wizard/step3');
            } else {
                $data['error'] = 'Database Error: ' . mysql_error();
            }
        }
        $this->_output_step('steps/step2', $data);
    }

    public function step3() {
        $valid = TRUE;
        if ($this->session->userdata('last_step') != '2') {
            redirect('wizard/step2');
        }
        $this->load->library('form_validation');

        $this->form_validation->set_rules('site_name', 'Site Name', 'required|xss_clean');
        $this->form_validation->set_rules('license_key', 'License Key', 'required|xss_clean');
        $this->form_validation->set_rules('first_name', 'First Name', 'required|xss_clean');
        $this->form_validation->set_rules('last_name', 'Last Name', 'required|xss_clean');
        $this->form_validation->set_rules('username', 'Username', 'required|xss_clean');
        $this->form_validation->set_rules('notify_email', 'Notify Email', 'required|valid_email');
        $this->form_validation->set_rules('password', 'Password', 'required|min_length[5]|matches[password_confirm]');
        $this->form_validation->set_rules('password_confirm', 'Password Confirmation', 'required');
        $this->form_validation->set_rules('mailing_address', 'Mailing Address', '');
        $this->form_validation->set_rules('tax_rate', 'Tax Rate', '');
        $this->form_validation->set_rules('currency', 'Currency', '');

        if ($this->form_validation->run() AND ( $valid = $this->_check_license($this->input->post('license_key')))) {
            $this->session->set_userdata('install_data', $_POST);
            $this->session->set_userdata('last_step', '3');
            redirect('wizard/complete');
        }

        include FCPATH . 'system/pancake/config/currency.php';

        foreach ($config['currencies'] as $code => $currency) {
            $data['currencies'][$code] = $currency['name'] . ' (' . $code . ')';
        }

        if (!$valid) {
            $data['messages']['error'] = 'Invalid License Key. If you believe this is incorrect, please contact Support.';
        }
        $this->_output_step('steps/step3', $data);
    }

    public function complete() {
        if ($this->session->userdata('last_step') != '3') {
            redirect('wizard/step3');
        }

        $this->load->library('installer');

        $data = $this->session->userdata('install_data');
        $data['salt'] = substr(md5(uniqid(rand(), true)), 0, 10);
        $data['password'] = sha1($data['password'] . $data['salt']);

        // This will ensure they won't have to redo the entire install on an error
        if ($this->installer->install($data)) {
            $this->session->set_userdata('last_step', '0');
        }

        $this->_output_step('steps/complete');
    }

    public function error() {
        $this->_output_step('error');
    }

    private function _check_license($key) {
        require_once FCPATH . 'system/pancake/helpers/pancake_helper.php';
        $result = get_url_contents(MANAGE_PANCAKE_BASE_URL . 'verify/key/' . $key, false);
        if (empty($result)) {
            show_error('Pancake could not verify if the key "' . $key . '" is valid.');
        }

        return $result === 'valid';
    }

    private function _output_step($view, $data = array()) {
        $content = $this->load->view($view, $data, TRUE);

        $this->load->view('template', array('content' => $content));
    }

}

/* End of file wizard.php */